########################################################################
# File name: utils.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################

import operator


def xor_bytes(a: bytes, b: bytes) -> bytes:
    """
    Calculate the byte wise exclusive of of two :class:`bytes` objects
    of the same length.
    """
    assert len(a) == len(b)
    return bytes(map(operator.xor, a, b))



########################################################################
# File name: channel_binding.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
Channel binding methods
=======================

The module :mod:`aiosasl.channel_binding` provides
implementations of the :class:`~.ChannelBindingProvider`
interface for use with :mod:`ssl` respective :mod:`OpenSSL`.

.. autoclass:: ChannelBindingProvider

.. autoclass:: StdlibTLS

.. autoclass:: TLSUnique

.. autoclass:: TLSServerEndPoint
"""
import abc
import functools
import ssl

try:
    import OpenSSL  # for mypy
except ImportError:
    pass


class ChannelBindingProvider(metaclass=abc.ABCMeta):
    """
    Interface for a channel binding method.

    The needed external information is supplied to the constructors of
    the specific instances.
    """

    @abc.abstractproperty
    def cb_name(self) -> bytes:
        """
        Return the name of the channel-binding mechanism.
        :rtype: :class:`bytes`
        """
        raise NotImplementedError

    @abc.abstractmethod
    def extract_cb_data(self) -> bytes:
        """
        Return the channel binding data.
        :returns: the channel binding data
        :rtype: :class:`bytes`
        """
        raise NotImplementedError


class StdlibTLS(ChannelBindingProvider):
    """
    Provider for channel binding for :mod:`ssl`.

    :param connection: the SSL connection
    :type connection: :class:`ssl.SSLSocket`
    :param type_: the channel binding type
    :type type_: :class:`str`
    """

    def __init__(
            self,
            connection: ssl.SSLSocket,
            type_: str):
        super().__init__()
        self._connection = connection
        self._type = type_

    @property
    def cb_name(self) -> bytes:
        return self._type.encode("us-ascii")

    def extract_cb_data(self) -> bytes:
        return self._connection.get_channel_binding(self._type)  # type:ignore


class TLSUnique(ChannelBindingProvider):
    """
    Provider for the channel binding ``tls-unique`` as specified by
    :rfc:`5929` for :mod:`OpenSSL`.

    .. warning::

       This only supports connections that were not created by session
       resumption.

    :param connection: the SSL connection
    :type connection: :class:`OpenSSL.SSL.Connection`
    """

    def __init__(self, connection: "OpenSSL.SSL.Connection"):
        super().__init__()
        self._connection = connection

    @property
    def cb_name(self) -> bytes:
        return b"tls-unique"

    def extract_cb_data(self) -> bytes:
        return self._connection.get_finished()


def parse_openssl_digest(
        digest: bytes,
        ) -> bytes:
    return bytes(map(functools.partial(int, base=16), digest.split(b":")))


class TLSServerEndPoint(ChannelBindingProvider):
    """
    Provider for the channel binding ``tls-server-end-point`` as
    specified by :rfc:`5929` for :mod:`OpenSSL`.

    :param connection: the SSL connection
    :type connection: :class:`OpenSSL.SSL.Connection`
    """

    def __init__(
            self,
            connection: "OpenSSL.SSL.Connection"):
        super().__init__()
        self._connection = connection

    @property
    def cb_name(self) -> bytes:
        return b"tls-server-end-point"

    def extract_cb_data(self) -> bytes:
        cert = self._connection.get_peer_certificate()
        algo, part, _ = cert.get_signature_algorithm().lower().partition(
            b"with")
        if not part:
            raise NotImplementedError
        if algo in (b"sha1", b"md5"):
            algo = b"sha256"
        return parse_openssl_digest(cert.digest(algo.decode("us-ascii")))



########################################################################
# File name: scram.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import base64
import collections
import functools
import hashlib
import hmac
import logging
import random
import time
import typing

from hashlib import pbkdf2_hmac as pbkdf2

from . import channel_binding, common, statemachine, stringprep, utils


logger = logging.getLogger(__name__)


SCRAMHashInfo = collections.namedtuple(
    "SCRAMHashInfo",
    [
        "hashfun_name",
        "quality",
        "minimum_iteration_count",
    ]
)


_system_random = random.SystemRandom()


class Base:
    """
    Shared implementation of SCRAM and SCRAMPLUS.
    """

    _channel_binding = False

    def __init__(
            self,
            credential_provider: common.CredentialProvider,
            *,
            nonce_length: int = 15,
            enforce_minimum_iteration_count: bool = True):
        super().__init__()
        self._credential_provider = credential_provider
        self.nonce_length = nonce_length
        self.enforce_minimum_iteration_count = enforce_minimum_iteration_count

    _supported_hashalgos = {
        # the second argument is for preference ordering (highest first)
        # if anyone has a better hash ordering suggestion, I’m open for it
        # a value of 1 is added if the -PLUS variant is used
        # -- JSC
        # the minimum iteration count is obtained from
        # <https://www.iana.org/assignments/sasl-mechanisms/sasl-mechanisms.xhtml>
        "SHA-1": SCRAMHashInfo("sha1", 1, 4096),
        "SHA-256": SCRAMHashInfo("sha256", 256, 4096),
    }

    @classmethod
    def any_supported(
            cls,
            mechanisms: typing.Iterable[str],
            ) -> typing.Optional[typing.Tuple[str, SCRAMHashInfo]]:
        supported = []
        for mechanism in mechanisms:
            if not mechanism.startswith("SCRAM-"):
                continue

            hashfun_key = mechanism[6:]

            if cls._channel_binding:
                if not mechanism.endswith("-PLUS"):
                    continue
                hashfun_key = hashfun_key[:-5]
            else:
                if mechanism.endswith("-PLUS"):
                    continue

            try:
                info = cls._supported_hashalgos[hashfun_key]
            except KeyError:
                continue

            supported.append(((1, info.quality), (mechanism, info,)))

        if not supported:
            return None
        supported.sort()

        return supported.pop()[1]

    @classmethod
    def parse_message(
            cls,
            msg: bytes,
            ) -> typing.Generator[typing.Tuple[bytes, bytes], None, None]:
        parts = (
            part
            for part in msg.split(b",")
            if part)

        for part in parts:
            key, _, value = part.partition(b"=")
            if len(key) > 1 or key == b"m":
                raise Exception("SCRAM protocol violation / unknown "
                                "future extension")
            if key == b"n" or key == b"a":
                value = value.replace(b"=2C", b",").replace(b"=3D", b"=")

            yield key, value

    @abc.abstractmethod
    def _get_gs2_header(self) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_cb_data(self) -> bytes:
        raise NotImplementedError

    async def authenticate(
            self,
            sm: statemachine.SASLStateMachine,
            token: typing.Tuple[str, SCRAMHashInfo],
            ) -> None:
        mechanism, info, = token
        logger.info("attempting %s mechanism (using %s hashfun)",
                    mechanism,
                    info)
        # this is pretty much a verbatim implementation of RFC 5802.

        hashfun_factory = functools.partial(hashlib.new, info.hashfun_name)

        gs2_header = self._get_gs2_header()
        username, password = await self._credential_provider()
        encoded_username = stringprep.saslprep(
            username,
            allow_unassigned=True,
        ).encode("utf-8")
        encoded_password = stringprep.saslprep(password).encode("utf-8")

        our_nonce = base64.b64encode(_system_random.getrandbits(
            self.nonce_length * 8
        ).to_bytes(
            self.nonce_length, "little"
        ))

        auth_message = b"n=" + encoded_username + b",r=" + our_nonce
        state, payload = await sm.initiate(
            mechanism,
            gs2_header + auth_message)

        if state != common.SASLState.CHALLENGE or payload is None:
            await sm.abort()
            raise common.SASLFailure(
                None,
                text="protocol violation: expected challenge with payload")

        auth_message += b"," + payload

        parsed_payload = dict(self.parse_message(payload))

        try:
            iteration_count = int(parsed_payload[b"i"])
            nonce = parsed_payload[b"r"]
            salt = base64.b64decode(parsed_payload[b"s"])
        except (ValueError, KeyError):
            await sm.abort()
            raise common.SASLFailure(
                None,
                text="malformed server message: {!r}".format(payload),
            )

        if not nonce.startswith(our_nonce):
            await sm.abort()
            raise common.SASLFailure(
                None,
                text="server nonce doesn't fit our nonce")

        if (self.enforce_minimum_iteration_count and
                iteration_count < info.minimum_iteration_count):
            raise common.SASLFailure(
                None,
                text="minimum iteration count for {} violated "
                "({} is less than {})".format(
                    mechanism,
                    iteration_count,
                    info.minimum_iteration_count,
                )
            )

        t0 = time.time()

        salted_password = pbkdf2(
            info.hashfun_name,
            encoded_password,
            salt,
            iteration_count)

        logger.debug("pbkdf2 timing: %f seconds", time.time() - t0)

        client_key = hmac.new(
            salted_password,
            b"Client Key",
            hashfun_factory).digest()

        stored_key = hashfun_factory(client_key).digest()

        reply = b"c=" + base64.b64encode(self._get_cb_data()) + b",r=" + nonce

        auth_message += b"," + reply

        client_proof = utils.xor_bytes(
            hmac.new(
                stored_key,
                auth_message,
                hashfun_factory).digest(),
            client_key)

        logger.debug("response generation time: %f seconds", time.time() - t0)
        try:
            state, payload = await sm.response(
                reply + b",p=" + base64.b64encode(client_proof)
            )
        except common.SASLFailure as err:
            raise err.promote_to_authentication_failure() from None

        # this is the pseudo-challenge for the server signature
        # we have to reply with the empty string!
        if state != common.SASLState.CHALLENGE:
            raise common.SASLFailure(
                "malformed-request",
                text="SCRAM protocol violation")

        state, dummy_payload = await sm.response(b"")
        if state != common.SASLState.SUCCESS or dummy_payload is not None:
            raise common.SASLFailure(
                None,
                "SASL protocol violation")

        server_signature = hmac.new(
            hmac.new(
                salted_password,
                b"Server Key",
                hashfun_factory).digest(),
            auth_message,
            hashfun_factory).digest()

        parsed_payload = dict(self.parse_message(payload or b""))

        if base64.b64decode(parsed_payload[b"v"]) != server_signature:
            raise common.SASLFailure(
                None,
                "authentication successful, but server signature invalid",
            )


class SCRAM(Base, statemachine.SASLMechanism):
    """
    The password-based SCRAM (non-PLUS) SASL mechanism (see :rfc:`5802`).

    :param credential_provider: A coroutine function which returns credentials.
    :param after_scram_plus: Flag to indicate that SCRAM-PLUS *is* supported by
        your implementation.
    :type after_scram_plus: :class:`bool`
    :param enforce_minimum_iteration_count: Enforce the minimum iteration
        count specified by the SCRAM specifications.
    :type enforce_minimum_iteration_count: :class:`bool`

    .. note::

       As "non-PLUS" suggests, this does not support channel binding.
       Use :class:`SCRAMPLUS` if you want channel binding.


    `credential_provider` must be coroutine function which returns a ``(user,
    password)`` tuple.

    If this is used after :class:`SCRAMPLUS` in a method list, the
    keyword argument `after_scram_plus` should be set to
    :data:`True`. Then we will use the gs2 header ``y,,`` to prevent
    down-grade attacks by a man-in-the-middle attacker.

    `enforce_minimum_iteration_count` controls the enforcement of the specified
    minimum iteration count for the key derivation function used in SCRAM. By
    default, this enforcement is enabled, and you are strongly advised to not
    disable it: it can be used to make the exchange weaker.

    Disabling `enforce_minimum_iteration_count` only makes sense if the
    authentication exchange would otherwise fall back to using :class:`PLAIN`
    or a similarly weak authentication mechanism.

    .. versionchanged:: 0.4

        The `enforce_minimum_iteration_count` argument and the behaviour to
        enforce the minimum iteration count by default was added.
    """

    def __init__(
            self,
            credential_provider: common.CredentialProvider,
            *,
            after_scram_plus: bool = False,
            **kwargs: typing.Any):
        super().__init__(credential_provider, **kwargs)
        self._after_scram_plus = after_scram_plus

    def _get_gs2_header(self) -> bytes:
        if self._after_scram_plus:
            return b"y,,"
        else:
            return b"n,,"

    def _get_cb_data(self) -> bytes:
        return self._get_gs2_header()


class SCRAMPLUS(Base, statemachine.SASLMechanism):
    """
    The password-based SCRAM-PLUS SASL mechanism (see :rfc:`5802`).

    :param credential_provider: A coroutine function which returns credentials.
    :param cb_provider: Object which provides channel binding data and
        information.
    :type cb_provider: :class:`.ChannelBindingProvider`
    :param after_scram_plus: Flag to indicate that SCRAM-PLUS *is* supported by
        your implementation.
    :type after_scram_plus: :class:`bool`
    :param enforce_minimum_iteration_count: Enforce the minimum iteration
        count specified by the SCRAM specifications.
    :type enforce_minimum_iteration_count: :class:`bool`

    `credential_provider` must be coroutine which returns a ``(user,
    password)`` tuple.

    `cb_provider` must be an instance of
    :class:`.ChannelBindingProvider`, which specifies and implements
    the channel binding type to use.

    `enforce_minimum_iteration_count` controls the enforcement of the specified
    minimum iteration count for the key derivation function used in SCRAM. By
    default, this enforcement is enabled, and you are strongly advised to not
    disable it: it can be used to make the exchange weaker.

    .. seealso::

        :class:`SCRAM` for more information on
        `enforce_minimum_iteration_count`.

    .. versionchanged:: 0.4

        The `enforce_minimum_iteration_count` argument and the behaviour to
        enforce the minimum iteration count by default was added.

    """
    _channel_binding = True

    def __init__(self,
                 credential_provider: common.CredentialProvider,
                 cb_provider: channel_binding.ChannelBindingProvider,
                 **kwargs: typing.Any):
        super().__init__(credential_provider, **kwargs)
        self._cb_provider = cb_provider

    def _get_gs2_header(self) -> bytes:
        return b"p=" + self._cb_provider.cb_name + b",,"

    def _get_cb_data(self) -> bytes:
        gs2_header = self._get_gs2_header()
        cb_data = self._cb_provider.extract_cb_data()
        return gs2_header + cb_data



########################################################################
# File name: plain.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import logging
import typing

from . import common, statemachine


logger = logging.getLogger(__name__)


class PLAIN(statemachine.SASLMechanism):
    """
    The password-based ``PLAIN`` SASL mechanism (see :rfc:`4616`).

    .. warning::

       This is generally unsafe over unencrypted connections and should not be
       used there. Exclusion of the ``PLAIN`` mechanism over unsafe connections
       is out of scope for :mod:`aiosasl` and needs to be handled by the
       protocol implementation!

    `credential_provider` must be coroutine which returns a ``(user,
    password)`` tuple.
    """
    def __init__(self, credential_provider: common.CredentialProvider):
        super().__init__()
        self._credential_provider = credential_provider

    @classmethod
    def any_supported(
            cls,
            mechanisms: typing.Iterable[str],
            ) -> typing.Any:
        if "PLAIN" in mechanisms:
            return "PLAIN"
        return None

    async def authenticate(
            self,
            sm: statemachine.SASLStateMachine,
            mechanism: typing.Any,
            ) -> None:
        logger.info("attempting PLAIN mechanism")
        username, password = await self._credential_provider()
        encoded_username = username.encode("utf8")
        encoded_password = password.encode("utf8")

        if b"\0" in encoded_username or b"\0" in encoded_password:
            raise ValueError("NUL byte in username or password is disallowed")

        state, _ = await sm.initiate(
            mechanism="PLAIN",
            payload=b"\0" + encoded_username + b"\0" + encoded_password,
        )

        if state != common.SASLState.SUCCESS:
            raise common.SASLFailure(
                None,
                text="SASL protocol violation")



########################################################################
# File name: test_unicode_version.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################

import unittest

import aiosasl.stringprep


class TestUnicodeVersion(unittest.TestCase):
    def test_version(self):
        self.assertEqual(
            aiosasl.stringprep.unicodedata.unidata_version,
            "3.2.0",
        )



########################################################################
# File name: test_stringprep.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

from aiosasl.stringprep import (
    saslprep,
    check_bidi,
    trace,
)


class Testcheck_bidi(unittest.TestCase):
    # some test cases which are not covered by the other tests
    def test_empty_string(self):
        check_bidi("")

    def test_L_RAL_violation(self):
        with self.assertRaises(ValueError):
            check_bidi("\u05be\u0041")


class TestSASLprep(unittest.TestCase):
    def test_map_to_nothing_rfcx(self):
        self.assertEqual(
            "IX",
            saslprep("I\u00ADX"),
            "SASLprep requirement: map SOFT HYPHEN to nothing")

    def test_map_to_space(self):
        self.assertEqual(
            "I X",
            saslprep("I\u00A0X"),
            "SASLprep requirement: map SOFT HYPHEN to nothing")

    def test_identity_rfcx(self):
        self.assertEqual(
            "user",
            saslprep("user"),
            "SASLprep requirement: identity transform")

    def test_case_preservation_rfcx(self):
        self.assertEqual(
            "USER",
            saslprep("USER"),
            "SASLprep requirement: preserve case")

    def test_nfkc_rfcx(self):
        self.assertEqual(
            "a",
            saslprep("\u00AA"),
            "SASLprep requirement: NFKC")
        self.assertEqual(
            "IX",
            saslprep("\u2168"),
            "SASLprep requirement: NFKC")

    def test_prohibited_character_rfcx(self):
        with self.assertRaises(
                ValueError,
                msg="SASLprep requirement: prohibited character (C.2.1)"):
            saslprep("\u0007")

        with self.assertRaises(
                ValueError,
                msg="SASLprep requirement: prohibited character (C.8)"):
            saslprep("\u200E")

    def test_bidirectional_check_rfcx(self):
        with self.assertRaises(
                ValueError,
                msg="SASLprep requirement: bidirectional check"):
            saslprep("\u0627\u0031")

    def test_unassigned(self):
        with self.assertRaises(
                ValueError,
                msg="SASLprep requirement: unassigned"):
            saslprep("\u0221", allow_unassigned=False)

        with self.assertRaises(
                ValueError,
                msg="enforce no unassigned by default"):
            saslprep("\u0221")

        self.assertEqual(
            "\u0221",
            saslprep("\u0221", allow_unassigned=True))


class Testtrace(unittest.TestCase):
    def test_identity_rfcx(self):
        self.assertEqual(
            "user",
            trace("user"),
            "trace requirement: identity transform")

    def test_case_preservation_rfcx(self):
        self.assertEqual(
            "USER",
            trace("USER"),
            "trace requirement: preserve case")

    def test_prohibited_character_rfcx(self):
        with self.assertRaises(
                ValueError,
                msg="trace requirement: prohibited character (C.2.1)"):
            trace("\u0007")

        with self.assertRaises(
                ValueError,
                msg="trace requirement: prohibited character (C.2.2)"):
            trace("\u070F")

        with self.assertRaises(
                ValueError,
                msg="trace requirement: prohibited character (C.3)"):
            trace("\uE000")

        with self.assertRaises(
                ValueError,
                msg="trace requirement: prohibited character (C.4)"):
            trace("\uFDEF")

        with self.assertRaises(
                ValueError,
                msg="trace requirement: prohibited character (C.5)"):
            trace("\uD800")

        with self.assertRaises(
                ValueError,
                msg="trace requirement: prohibited character (C.6)"):
            trace("\uFFF9")

        with self.assertRaises(
                ValueError,
                msg="trace requirement: prohibited character (C.8)"):
            trace("\u0340")

        with self.assertRaises(
                ValueError,
                msg="trace requirement: prohibited character (C.9)"):
            trace("\U000E0001")

    def test_bidirectional_check_rfcx(self):
        with self.assertRaises(
                ValueError,
                msg="trace requirement: bidirectional check"):
            trace("\u0627\u0031")



########################################################################
# File name: test_channel_binding_methods.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################

import unittest
import unittest.mock

import OpenSSL

from aiosasl.channel_binding import (
    ChannelBindingProvider,
    StdlibTLS,
    TLSUnique,
    TLSServerEndPoint,
    parse_openssl_digest,
)


RAW_EXAMPLE_CERTS = [(b'0\x82\x01\xf90\x82\x01b\x02\x01\x010\r\x06\t*\x86H\x86\xf7\r\x01\x01\x04\x05\x000C1\x0b0\t\x06\x03U\x04\x06\x13\x02FN1\x1c0\x1a\x06\x03U\x04\n\x0c\x13Example Association1\x160\x14\x06\x03U\x04\x03\x0c\raiosasl tests0"\x18\x0f20000101000000Z\x18\x0f21000101000000Z0C1\x0b0\t\x06\x03U\x04\x06\x13\x02FN1\x1c0\x1a\x06\x03U\x04\n\x0c\x13Example Association1\x160\x14\x06\x03U\x04\x03\x0c\raiosasl tests0\x81\x9f0\r\x06\t*\x86H\x86\xf7\r\x01\x01\x01\x05\x00\x03\x81\x8d\x000\x81\x89\x02\x81\x81\x00\x9d\xce/J\x9b\x89Z{~g\xda\xd8\xbfCi%g:\x9f\xbb\xcb\xe6\x08\xd0\x9c\x0b\xec1\xda\xb8\xd1w\x14v\xc83\x8dI\xc7\xf1#qz\x97\xf7\xf27\xd2\x97\x91s\x1f/i\x06f\x1a\x1ejN\xfb\x13\xf1\xe1\x8a\'\xa6\xec\x03h\xdd+\xbe\xda\xde\xba\xfd\xe7\xc4h\xc6]5d\xe5k\x97\x15\xf1\xc4\xc0\xf1\xc8\xb7\xb4\xea\x0f\xbb\x15O\xc10 3\x8d\x9a \xa8Dg\xb1\x1c\xac\xa3\xe7pY\xdb\xb7\xb8Ze\x95^\x94\x1b\xfa\xed\x02\x03\x01\x00\x010\r\x06\t*\x86H\x86\xf7\r\x01\x01\x04\x05\x00\x03\x81\x81\x00l77\x1c\x95/\xb41G\xe8l\xe6\x9e\x8aL\xbcs#?\x8diZwB\x942\xa0\x82\x1e\xa5\x08\xe5%\xd8\x93<\xae\xe7\x1fR\x03\xf7\xcfw\xf4\xf7Q\x99B\x9c\xbf\x80\xac\xd5{\r\xa7\xf2\xb1\xfa\x88\xd2\x14\xf4\xf9q(\xfa4\x17\xa9\x07V\xb4\xe8G\xb4\x93\x8b\x8b\x0b\xc7\x00\xc5/\x80\x8c\x1d\x85Uv\xd5\xa9\xe6\x17\xb1m@\xb7\x01\xd6.l\xe6\xa9/\x90\xea$NB\xfa\xa5\xe3\xe4\xf4I\xfcQ\x85\x06\xc7\xd3g\xb8\xe9\xf5i', b'0B:AA:BA:DD:CA:F4:7C:2F:5A:2A:12:67:16:B3:1C:AF:65:1F:D6:B0:BF:A2:B7:B9:9A:CD:83:F4:AB:B5:CE:CD'), (b'0\x82\x01\xf90\x82\x01b\x02\x01\x010\r\x06\t*\x86H\x86\xf7\r\x01\x01\x05\x05\x000C1\x0b0\t\x06\x03U\x04\x06\x13\x02FN1\x1c0\x1a\x06\x03U\x04\n\x0c\x13Example Association1\x160\x14\x06\x03U\x04\x03\x0c\raiosasl tests0"\x18\x0f20000101000000Z\x18\x0f21000101000000Z0C1\x0b0\t\x06\x03U\x04\x06\x13\x02FN1\x1c0\x1a\x06\x03U\x04\n\x0c\x13Example Association1\x160\x14\x06\x03U\x04\x03\x0c\raiosasl tests0\x81\x9f0\r\x06\t*\x86H\x86\xf7\r\x01\x01\x01\x05\x00\x03\x81\x8d\x000\x81\x89\x02\x81\x81\x00\x9d\xce/J\x9b\x89Z{~g\xda\xd8\xbfCi%g:\x9f\xbb\xcb\xe6\x08\xd0\x9c\x0b\xec1\xda\xb8\xd1w\x14v\xc83\x8dI\xc7\xf1#qz\x97\xf7\xf27\xd2\x97\x91s\x1f/i\x06f\x1a\x1ejN\xfb\x13\xf1\xe1\x8a\'\xa6\xec\x03h\xdd+\xbe\xda\xde\xba\xfd\xe7\xc4h\xc6]5d\xe5k\x97\x15\xf1\xc4\xc0\xf1\xc8\xb7\xb4\xea\x0f\xbb\x15O\xc10 3\x8d\x9a \xa8Dg\xb1\x1c\xac\xa3\xe7pY\xdb\xb7\xb8Ze\x95^\x94\x1b\xfa\xed\x02\x03\x01\x00\x010\r\x06\t*\x86H\x86\xf7\r\x01\x01\x05\x05\x00\x03\x81\x81\x00\x92$\xfe\xc8\xeb~#\xd5\x0f#\x1c\xd9\xfeV\xa4z\x84\xcc\x83\xdd\xfd\xa47\xb5\xf7\xe4*\x9fJ\xd5My\x13UFT+\x95\'v\xa8\xb6\x95{2\xe4t\xbd#F\x08\xbf\xf3\xbei\xaf\x83M\xbb\x06FO\x1b\x9ew\xc81\r\xa0\xc3Z\xb8\xb7\x16\tS\xde\xe8\xe4\xdd\xed\x04\xfb\xe1\x1a(\xf0i}\x18\x10Q\x82\x10\x8eH\xe9x:\xcd\xaec]\x9c\xd9\xa7&\xf1\xa3:`\xa8(\x86+\xf05\n!\x82\xac\xfd\xd7\xe72T\x15.', b'79:45:47:63:08:2F:E0:21:E4:31:A5:99:EE:FE:34:D5:9E:1A:A9:A0:54:00:D6:FE:E8:89:62:27:A1:8E:5F:A7'), (b'0\x82\x01\xf90\x82\x01b\x02\x01\x010\r\x06\t*\x86H\x86\xf7\r\x01\x01\x0b\x05\x000C1\x0b0\t\x06\x03U\x04\x06\x13\x02FN1\x1c0\x1a\x06\x03U\x04\n\x0c\x13Example Association1\x160\x14\x06\x03U\x04\x03\x0c\raiosasl tests0"\x18\x0f20000101000000Z\x18\x0f21000101000000Z0C1\x0b0\t\x06\x03U\x04\x06\x13\x02FN1\x1c0\x1a\x06\x03U\x04\n\x0c\x13Example Association1\x160\x14\x06\x03U\x04\x03\x0c\raiosasl tests0\x81\x9f0\r\x06\t*\x86H\x86\xf7\r\x01\x01\x01\x05\x00\x03\x81\x8d\x000\x81\x89\x02\x81\x81\x00\x9d\xce/J\x9b\x89Z{~g\xda\xd8\xbfCi%g:\x9f\xbb\xcb\xe6\x08\xd0\x9c\x0b\xec1\xda\xb8\xd1w\x14v\xc83\x8dI\xc7\xf1#qz\x97\xf7\xf27\xd2\x97\x91s\x1f/i\x06f\x1a\x1ejN\xfb\x13\xf1\xe1\x8a\'\xa6\xec\x03h\xdd+\xbe\xda\xde\xba\xfd\xe7\xc4h\xc6]5d\xe5k\x97\x15\xf1\xc4\xc0\xf1\xc8\xb7\xb4\xea\x0f\xbb\x15O\xc10 3\x8d\x9a \xa8Dg\xb1\x1c\xac\xa3\xe7pY\xdb\xb7\xb8Ze\x95^\x94\x1b\xfa\xed\x02\x03\x01\x00\x010\r\x06\t*\x86H\x86\xf7\r\x01\x01\x0b\x05\x00\x03\x81\x81\x000\xbfiz\x97\x90\xab\x8f\xca1\x1eH\xeb\xd6\xbd?\x87o\x07\xd3\xddi\xe9C\xa9\x97\x84\x87Xp&&\'\x0c\xbci\xc3\xb4?o\xfdOy\x1d\xf2\x9e\xe36\x8d(\xc2A\x0c\x106$}\xb3X;\xa7\xc3h\xd38D\xeb\x95\x8f>6\x17\xa1e\x1a\xc9\xfa\xbd\r\x00\x8c0\xa5T{n5\xe5\x9dp\x80%\x9eb\xb4\xd6\xd3;\xc7\x8c\n\x19\x9b\xaf\xab =IhK\xfc%*\xdbMkM\x8f\x1d\x05\xd8\xa9\xbd\x17RW\xfc{', b'0A:19:C9:73:FE:9C:F2:B9:DF:5D:27:CC:5A:FD:04:5E:19:97:05:99:4A:EB:91:16:FC:3F:CC:87:1B:D5:6E:2E'), (b'0\x82\x01\xf90\x82\x01b\x02\x01\x010\r\x06\t*\x86H\x86\xf7\r\x01\x01\r\x05\x000C1\x0b0\t\x06\x03U\x04\x06\x13\x02FN1\x1c0\x1a\x06\x03U\x04\n\x0c\x13Example Association1\x160\x14\x06\x03U\x04\x03\x0c\raiosasl tests0"\x18\x0f20000101000000Z\x18\x0f21000101000000Z0C1\x0b0\t\x06\x03U\x04\x06\x13\x02FN1\x1c0\x1a\x06\x03U\x04\n\x0c\x13Example Association1\x160\x14\x06\x03U\x04\x03\x0c\raiosasl tests0\x81\x9f0\r\x06\t*\x86H\x86\xf7\r\x01\x01\x01\x05\x00\x03\x81\x8d\x000\x81\x89\x02\x81\x81\x00\x9d\xce/J\x9b\x89Z{~g\xda\xd8\xbfCi%g:\x9f\xbb\xcb\xe6\x08\xd0\x9c\x0b\xec1\xda\xb8\xd1w\x14v\xc83\x8dI\xc7\xf1#qz\x97\xf7\xf27\xd2\x97\x91s\x1f/i\x06f\x1a\x1ejN\xfb\x13\xf1\xe1\x8a\'\xa6\xec\x03h\xdd+\xbe\xda\xde\xba\xfd\xe7\xc4h\xc6]5d\xe5k\x97\x15\xf1\xc4\xc0\xf1\xc8\xb7\xb4\xea\x0f\xbb\x15O\xc10 3\x8d\x9a \xa8Dg\xb1\x1c\xac\xa3\xe7pY\xdb\xb7\xb8Ze\x95^\x94\x1b\xfa\xed\x02\x03\x01\x00\x010\r\x06\t*\x86H\x86\xf7\r\x01\x01\r\x05\x00\x03\x81\x81\x00\x8cf\x94\xeet\xf4\x03\xe6Tj|\xee\x1dh?\xed\x9b4\xb4\xb1\xc0J\xa6\xe1\xb8$\xf9c)\xd0[5\xed\x8d\xa3 \x9f\xfb\xedm\x904\x9a(u\xbe\x0b\xa8\\)d\xb2\x8b\xd6\xf2^\x80\xa4Z\x10\x0b\x8aN\'f\xbb\x81\x95\xc2\x99v\x96\xb2\xb5_\xed\xcc\xc1\x9a\xc1\xa7\x85\x7f\xa3s\x17\xab\x98\x91o\xdd3\xd64{\x9c\xaft3t`\xa8\'\x9b\x7f\x13\x02.\xf2Vl=3\xd5#\x8e>L\x8c\xfe\xce\x1c\xb3\x7f\x04\xe8m\xad', b'B1:90:A4:38:8E:0E:63:F4:0E:3B:D3:21:8F:C5:33:2E:3B:96:F7:86:14:C1:E1:37:ED:AE:BE:B1:4C:96:37:7A:E7:13:83:4D:94:B6:E8:2C:D8:A6:E0:3C:DA:AA:79:07:2A:73:85:58:C9:D1:ED:CE:53:E0:A8:F7:B1:A0:DB:62')]  # NOQA


EXAMPLE_CERTS = []
for cert, digest in RAW_EXAMPLE_CERTS:
    EXAMPLE_CERTS.append((
        OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_ASN1, cert),
        parse_openssl_digest(digest),
    ))


class TestStdlibTLS(unittest.TestCase):

    def test_is_channel_binding_type(self):
        self.assertTrue(issubclass(StdlibTLS, ChannelBindingProvider))

    def test_cb_name(self):
        for method in ("tls-unique", "tls-server-end-point"):
            mock_socket = unittest.mock.Mock()
            provider = StdlibTLS(mock_socket, method)
            self.assertEqual(provider.cb_name, method.encode("us-ascii"))

        mock_socket = unittest.mock.Mock()
        provider = StdlibTLS(mock_socket, "tls-server-end-point")
        self.assertEqual(provider.cb_name, b"tls-server-end-point")

    def test_extract_cb_data(self):
        for method in ("tls-unique", "tls-server-end-point"):
            mock_socket = unittest.mock.Mock()
            provider = StdlibTLS(mock_socket, method)
            with unittest.mock.patch.object(
                    mock_socket,
                    "get_channel_binding") as get_channel_binding:
                get_channel_binding.return_value = b"foobar"
                cb_data = provider.extract_cb_data()

            self.assertSequenceEqual(
                get_channel_binding.mock_calls,
                [
                    unittest.mock.call(method)
                ]
            )

            self.assertEqual(cb_data, b"foobar")


class TestTLSUnique(unittest.TestCase):

    def test_is_channel_binding_type(self):
        self.assertTrue(issubclass(TLSUnique, ChannelBindingProvider))

    def test_cb_name(self):
        mock_conn = unittest.mock.Mock()
        provider = TLSUnique(mock_conn)
        self.assertEqual(provider.cb_name, b"tls-unique")

    def test_extract_cb_data(self):
        mock_conn = unittest.mock.Mock()
        provider = TLSUnique(mock_conn)
        with unittest.mock.patch.object(
                mock_conn,
                "get_finished") as get_finished:
            get_finished.return_value = b"foobar"
            cb_data = provider.extract_cb_data()

        self.assertSequenceEqual(
            get_finished.mock_calls,
            [
                unittest.mock.call()
            ]
        )

        self.assertEqual(cb_data, b"foobar")


class TestTLSServerEndPoint(unittest.TestCase):

    def test_is_channel_binding_type(self):
        self.assertTrue(issubclass(TLSServerEndPoint, ChannelBindingProvider))

    def test_cb_name(self):
        mock_conn = unittest.mock.Mock()
        provider = TLSServerEndPoint(mock_conn)
        self.assertEqual(provider.cb_name, b"tls-server-end-point")

    def test_extract_cb_data(self):
        mock_conn = unittest.mock.Mock()
        provider = TLSServerEndPoint(mock_conn)

        for cert, hash_ in EXAMPLE_CERTS:
            with unittest.mock.patch.object(
                    mock_conn,
                    "get_peer_certificate") as get_peer_cert:
                get_peer_cert.return_value = cert
                cb_data = provider.extract_cb_data()

            self.assertSequenceEqual(
                get_peer_cert.mock_calls,
                [
                    unittest.mock.call()
                ]
            )

            self.assertEqual(cb_data, hash_)



########################################################################
# File name: anonymous.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import logging
import typing

from . import common, statemachine, stringprep


logger = logging.getLogger()


class ANONYMOUS(statemachine.SASLMechanism):
    """
    The ANONYMOUS SASL mechanism (see :rfc:`4505`).

    .. versionadded:: 0.3
    """

    def __init__(self, token: str) -> None:
        super().__init__()
        self._token = stringprep.trace(token).encode("utf-8")

    @classmethod
    def any_supported(
            self,
            mechanisms: typing.Iterable[str],
            ) -> typing.Optional[str]:
        if "ANONYMOUS" in mechanisms:
            return "ANONYMOUS"
        return None

    async def authenticate(
            self,
            sm: statemachine.SASLStateMachine,
            mechanism: typing.Any) -> None:
        logger.info("attempting ANONYMOUS mechanism")

        state, _ = await sm.initiate(
            mechanism="ANONYMOUS",
            payload=self._token
        )

        if state != common.SASLState.SUCCESS:
            raise common.SASLFailure(
                None,
                text="SASL protocol violation")



########################################################################
# File name: common.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import enum
import typing


class SASLError(Exception):
    """
    Base class for a SASL related error. `opaque_error` may be anything but
    :data:`None` which helps your application re-identify the error at the
    outer layers. `kind` is a string which helps identifying the class of the
    error; this is set implicitly by the constructors of :class:`SASLFailure`
    and :class:`AuthenticationFailure`, which you are encouraged to use.

    `text` may be a human-readable string describing the error condition in
    more detail.

    `opaque_error` is set to :data:`None` by :class:`SASLMechanism`
    implementations to indicate errors which originate from the local mechanism
    implementation.

    .. attribute:: opaque_error

       The value passed to the respective constructor argument.

    .. attribute:: text

       The value passed to the respective constructor argument.

    """

    def __init__(
            self,
            opaque_error: typing.Any,
            kind: str,
            text: typing.Optional[str] = None):
        msg = "{}: {}".format(opaque_error, kind)
        if text:
            msg += ": {}".format(text)
        super().__init__(msg)
        self.opaque_error = opaque_error
        self.text = text


class AuthenticationFailure(SASLError):
    """
    A SASL error which indicates that the provided credentials are
    invalid. This may be raised by :class:`SASLInterface` methods.
    """

    def __init__(
            self,
            opaque_error: typing.Any,
            text: typing.Optional[str] = None):
        super().__init__(opaque_error, "authentication failed", text=text)


class SASLFailure(SASLError):
    """
    A SASL protocol failure which is unrelated to the credentials passed. This
    may be raised by :class:`SASLInterface` methods.
    """

    def __init__(
            self,
            opaque_error: typing.Any,
            text: typing.Optional[str] = None):
        super().__init__(opaque_error, "SASL failure", text=text)

    def promote_to_authentication_failure(self) -> AuthenticationFailure:
        return AuthenticationFailure(
            self.opaque_error,
            self.text)


class SASLState(enum.Enum):
    """
    The states of the SASL state machine.

    .. attribute:: CHALLENGE

       the server sent a SASL challenge

    .. attribute:: SUCCESS

       the authentication was successful

    .. attribute:: FAILURE

       the authentication failed

    Internal states used by the state machine:

    .. attribute:: INITIAL

       the state of the state machine before the
       authentication is started

    .. attribute:: SUCCESS_SIMULATE_CHALLENGE

       used to unwrap success replies that carry final data

    These internal states *must not* be returned by the
    :class:`SASLInterface` methods as first component of the result
    tuple.

    The following method is used to process replies returned
    by the :class:`SASLInterface` methods:

    .. method:: from_reply
    """

    INITIAL = "initial"
    CHALLENGE = "challenge"
    SUCCESS = "success"
    FAILURE = "failure"
    SUCCESS_SIMULATE_CHALLENGE = "success-simulate-challenge"

    @classmethod
    def from_reply(cls, state: "SASLState") -> "SASLState":
        """
        Comptaibility layer for old :class:`SASLInterface`
        implementations.

        Accepts the follwing set of :class:`SASLState` or strings and
        maps the strings to :class:`SASLState` elements as follows:

          ``"challenge"``
            :member:`SASLState.CHALLENGE`

           ``"failue"``
             :member:`SASLState.FAILURE`

           ``"success"``
             :member:`SASLState.SUCCESS`
        """
        if state in (SASLState.FAILURE, SASLState.SUCCESS,
                     SASLState.CHALLENGE):
            return state

        if state in ("failure", "success", "challenge"):
            return SASLState(state)
        else:
            raise RuntimeError("invalid SASL state", state)


NextStateTuple = typing.Tuple[SASLState, typing.Optional[bytes]]

CredentialProvider = typing.Callable[
    [], typing.Coroutine[typing.Any, typing.Any, typing.Tuple[str, str]]
]



########################################################################
# File name: test_aiosasl.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import base64
import hashlib
import hmac
import unittest
import unittest.mock

import aiosasl
import aiosasl.scram

from aiosasl.channel_binding import TLSUnique
from aiosasl.utils import xor_bytes


def run_coroutine(coroutine, timeout=1.0, loop=None):
    if not loop:
        loop = asyncio.get_event_loop()
    return loop.run_until_complete(
        asyncio.wait_for(
            coroutine,
            timeout=timeout))


class CoroutineMock(unittest.mock.Mock):
    delay = 0

    @asyncio.coroutine
    def __call__(self, *args, **kwargs):
        result = super().__call__(*args, **kwargs)
        yield from asyncio.sleep(self.delay)
        return result


class SASLInterfaceMock(aiosasl.SASLInterface):
    def __init__(self, testobj, action_sequence):
        super().__init__()
        self._testobj = testobj
        self._action_sequence = action_sequence

    def _check_action(self, action, payload):
        try:
            (next_action,
             next_payload,
             new_state,
             result_payload) = self._action_sequence.pop(0)
        except ValueError:
            raise AssertionError(
                "SASL action performed unexpectedly: "
                "{} with payload {}".format(
                    action,
                    payload))

        self._state = new_state

        self._testobj.assertEqual(
            action,
            next_action,
            "SASL action sequence violated")

        self._testobj.assertEqual(
            payload,
            next_payload,
            "SASL payload expectation violated")

        if new_state == "failure" and action != "abort":
            opaque_error, text = result_payload
            raise aiosasl.SASLFailure(opaque_error, text=text)

        if result_payload is not None:
            result_payload = result_payload

        return new_state, result_payload

    async def initiate(self, mechanism, payload=None):
        return self._check_action("auth;"+mechanism, payload)

    async def respond(self, payload):
        return self._check_action("response", payload)

    async def abort(self):
        return self._check_action("abort", None)

    def finalize(self):
        self._testobj.assertFalse(
            self._action_sequence,
            "Not all actions performed")


class TestSASLState(unittest.TestCase):

    def test_from_reply(self):
        self.assertEqual(
            aiosasl.SASLState.from_reply("success"),
            aiosasl.SASLState.SUCCESS
        )

        self.assertEqual(
            aiosasl.SASLState.from_reply("failure"),
            aiosasl.SASLState.FAILURE
        )

        self.assertEqual(
            aiosasl.SASLState.from_reply("challenge"),
            aiosasl.SASLState.CHALLENGE
        )

        self.assertEqual(
            aiosasl.SASLState.from_reply(aiosasl.SASLState.SUCCESS),
            aiosasl.SASLState.SUCCESS
        )

        self.assertEqual(
            aiosasl.SASLState.from_reply(aiosasl.SASLState.FAILURE),
            aiosasl.SASLState.FAILURE
        )

        self.assertEqual(
            aiosasl.SASLState.from_reply(aiosasl.SASLState.CHALLENGE),
            aiosasl.SASLState.CHALLENGE
        )

        with self.assertRaises(RuntimeError):
            aiosasl.SASLState.from_reply("initial"),

        with self.assertRaises(RuntimeError):
            aiosasl.SASLState.from_reply("success-simulate-initial"),

        with self.assertRaises(RuntimeError):
            aiosasl.SASLState.from_reply(aiosasl.SASLState.INITIAL),

        with self.assertRaises(RuntimeError):
            aiosasl.SASLState.from_reply(
                aiosasl.SASLState.SUCCESS_SIMULATE_CHALLENGE),


class TestSASLStateMachine(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.get_event_loop()
        self.intf = unittest.mock.Mock()
        self.intf.initiate = CoroutineMock()
        self.intf.respond = CoroutineMock()
        self.intf.abort = CoroutineMock()
        self.sm = aiosasl.SASLStateMachine(self.intf)

        self.intf.initiate.return_value = (aiosasl.SASLState.SUCCESS, None)

    def test_initiate_calls_to_interface(self):
        result = run_coroutine(
            self.sm.initiate("foo", b"bar")
        )

        self.intf.initiate.assert_called_with(
            "foo",
            payload=b"bar")

        self.assertEqual(
            run_coroutine(self.intf.initiate()),
            result
        )

    def test_reject_double_initiate(self):
        run_coroutine(self.sm.initiate("foo", b"bar"))

        with self.assertRaisesRegexp(RuntimeError,
                                     "has already been called"):
            run_coroutine(self.sm.initiate("foo"))

    def test_reject_double_initiate_after_error(self):
        opaque_error = object()
        self.intf.initiate.side_effect = aiosasl.SASLFailure(
            opaque_error
        )

        with self.assertRaises(aiosasl.SASLFailure):
            run_coroutine(self.sm.initiate("foo", b"bar"))

        with self.assertRaisesRegexp(RuntimeError,
                                     "has already been called"):
            run_coroutine(self.sm.initiate("foo"))

    def test_reject_response_without_challenge(self):
        with self.assertRaisesRegexp(RuntimeError,
                                     "no challenge"):
            run_coroutine(self.sm.response(b"bar"))

    def test_response_calls_to_interface(self):
        self.sm._state = aiosasl.SASLState.CHALLENGE
        self.intf.respond.return_value = (aiosasl.SASLState.SUCCESS, None)

        result = run_coroutine(
            self.sm.response(b"bar")
        )

        self.intf.respond.assert_called_with(b"bar")

        self.assertEqual(
            run_coroutine(self.intf.initiate()),
            result
        )

    def test_response_failure(self):
        opaque_error = object()
        self.sm._state = aiosasl.SASLState.CHALLENGE
        self.intf.respond.side_effect = aiosasl.SASLFailure(
            opaque_error
        )

        with self.assertRaises(aiosasl.SASLFailure):
            run_coroutine(
                self.sm.response(b"bar")
            )

        self.assertEqual(self.sm._state, aiosasl.SASLState.FAILURE)

    def test_reject_abort_without_initiate(self):
        with self.assertRaises(RuntimeError):
            run_coroutine(self.sm.abort())

    def test_abort_calls_to_interface(self):
        self.sm._state = "challenge"
        self.intf.abort.return_value = ("failure", None)

        self.assertEqual(
            ("failure", None),
            run_coroutine(self.sm.abort())
        )

        self.intf.abort.assert_called_with()
        self.assertEqual(self.sm._state, aiosasl.SASLState.FAILURE)

    def test_abort_set_to_failure_and_re_raise_exceptions(self):
        exc = Exception()
        self.sm._state = aiosasl.SASLState.CHALLENGE
        self.intf.abort.side_effect = exc

        with self.assertRaises(Exception) as ctx:
            run_coroutine(self.sm.abort())

        self.assertIs(ctx.exception, exc)

        self.intf.abort.assert_called_with()
        self.assertEqual(self.sm._state, aiosasl.SASLState.FAILURE)

    def test_success_simulated_challenge(self):
        self.sm._state = aiosasl.SASLState.CHALLENGE
        self.intf.respond.return_value = ("success", b"payload")
        state, payload = run_coroutine(self.sm.response(b"foobar"))
        self.assertEqual(self.sm._state,
                         aiosasl.SASLState.SUCCESS_SIMULATE_CHALLENGE)
        self.assertEqual(state, aiosasl.SASLState.CHALLENGE)
        self.assertEqual(payload, b"payload")
        state, payload = run_coroutine(self.sm.response(b""))
        self.assertEqual(state, aiosasl.SASLState.SUCCESS)
        self.assertEqual(payload, None)

    def test_success_simulated_challenge_protocol_violation(self):
        self.sm._state = aiosasl.SASLState.SUCCESS_SIMULATE_CHALLENGE
        with self.assertRaises(aiosasl.SASLFailure):
            run_coroutine(self.sm.response(b"not-empty"))
        self.assertEqual(self.sm._state, aiosasl.SASLState.FAILURE)

    def tearDown(self):
        del self.sm
        del self.intf
        del self.loop


class TestPLAIN(unittest.TestCase):
    def test_rfc(self):
        user = "tim"
        password = "tanstaaftanstaaf"

        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;PLAIN",
                 b"\0tim\0tanstaaftanstaaf",
                 "success",
                 None)
            ]))

        async def provide_credentials(*args):
            return user, password

        async def run():
            plain = aiosasl.PLAIN(provide_credentials)
            await plain.authenticate(
                smmock,
                "PLAIN",
            )

        asyncio.get_event_loop().run_until_complete(run())

        smmock.interface.finalize()

    def test_fail_on_protocol_violation(self):
        user = "tim"
        password = "tanstaaftanstaaf"

        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;PLAIN",
                 b"\0tim\0tanstaaftanstaaf",
                 "challenge",
                 b"foo")
            ]))

        async def provide_credentials(*args):
            return user, password

        async def run():
            plain = aiosasl.PLAIN(provide_credentials)
            await plain.authenticate(
                smmock,
                "PLAIN")

        with self.assertRaisesRegexp(aiosasl.SASLFailure,
                                     "protocol violation") as ctx:
            asyncio.get_event_loop().run_until_complete(run())

        self.assertEqual(
            None,
            ctx.exception.opaque_error
        )

        smmock.interface.finalize()

    def test_reject_NUL_bytes_in_username(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
            ]))

        async def provide_credentials(*args):
            return "\0", "foo"

        with self.assertRaises(ValueError):
            run_coroutine(
                aiosasl.PLAIN(provide_credentials).authenticate(
                    smmock,
                    "PLAIN")
            )

    def test_reject_NUL_bytes_in_password(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
            ]))

        async def provide_credentials(*args):
            return "foo", "\0"

        with self.assertRaises(ValueError):
            run_coroutine(
                aiosasl.PLAIN(provide_credentials).authenticate(
                    smmock,
                    "PLAIN",
                )
            )

    def test_does_not_apply_saslprep(self):
        user = "tim"
        password = "2ø'±s;ßà¼Å"

        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;PLAIN",
                 b"\0tim\0" + password.encode("utf-8"),
                 "success",
                 None)
            ]))

        async def provide_credentials(*args):
            return user, password

        async def run():
            plain = aiosasl.PLAIN(provide_credentials)
            await plain.authenticate(
                smmock,
                "PLAIN",
            )

        asyncio.get_event_loop().run_until_complete(run())

        smmock.interface.finalize()

    def test_supports_PLAIN(self):
        self.assertEqual(
            "PLAIN",
            aiosasl.PLAIN.any_supported(["PLAIN"])
        )

    def test_does_not_support_SCRAM(self):
        self.assertIsNone(
            aiosasl.PLAIN.any_supported(["SCRAM-SHA-1"])
        )


class TestSCRAMNegotiation(unittest.TestCase):
    def test_supports_SCRAM_famliy(self):
        hashes = ["SHA-1", "SHA-256"]

        for hashname in hashes:
            mechanism = "SCRAM-{}".format(hashname)
            self.assertEqual(
                (mechanism, unittest.mock.ANY),
                aiosasl.SCRAM.any_supported([mechanism])
            )

    def test_supports_SCRAMPLUS_famliy(self):
        hashes = ["SHA-1", "SHA-256"]

        for hashname in hashes:
            mechanism = "SCRAM-{}-PLUS".format(hashname)
            self.assertEqual(
                (mechanism, unittest.mock.ANY),
                aiosasl.SCRAMPLUS.any_supported([mechanism])
            )

    def test_pick_longest_hash_SCRAM(self):
        self.assertEqual(
            ("SCRAM-SHA-256", unittest.mock.ANY),
            aiosasl.SCRAM.any_supported([
                "SCRAM-SHA-1",
                "SCRAM-SHA-256",
                "PLAIN",
            ])
        )

    def test_no_support_for_unregistered_functions(self):
        self.assertEqual(
            ("SCRAM-SHA-256", unittest.mock.ANY),
            aiosasl.SCRAM.any_supported([
                "SCRAM-SHA-1",
                "SCRAM-SHA-256",
                "SCRAM-SHA-512",
                "PLAIN",
            ])
        )

    def test_pick_longest_hash_SCRAMPLUS(self):
        self.assertEqual(
            ("SCRAM-SHA-256-PLUS", unittest.mock.ANY),
            aiosasl.SCRAMPLUS.any_supported([
                "SCRAM-SHA-1-PLUS",
                "SCRAM-SHA-256-PLUS",
                "SCRAM-SHA-224-PLUS",
                "PLAIN",
            ])
        )

    def test_reject_scram_plus_SCRAM(self):
        hashes = ["SHA-1", "SHA-224", "SHA-256",
                  "SHA-512", "SHA-384", "SHA-256"]

        for hashname in hashes:
            mechanism = "SCRAM-{}-PLUS".format(hashname)
            self.assertIsNone(
                aiosasl.SCRAM.any_supported([mechanism])
            )

    def test_reject_scram_SCRAMPLUS(self):
        hashes = ["SHA-1", "SHA-256"]

        for hashname in hashes:
            mechanism = "SCRAM-{}".format(hashname)
            self.assertIsNone(
                aiosasl.SCRAMPLUS.any_supported([mechanism])
            )

    def test_reject_md5_SCRAM(self):
        self.assertIsNone(
            aiosasl.SCRAM.any_supported(["SCRAM-MD5"])
        )

    def test_reject_md5_SCRAMPLUS(self):
        self.assertIsNone(
            aiosasl.SCRAMPLUS.any_supported(["SCRAM-MD5-PLUS"])
        )

    def test_reject_unknown_hash_functions_SCRAM(self):
        self.assertIsNone(
            aiosasl.SCRAM.any_supported(["SCRAM-FOOBAR"])
        )

    def test_reject_unknown_hash_functions_SCRAMPLUS(self):
        self.assertIsNone(
            aiosasl.SCRAM.any_supported(["SCRAM-FOOBAR-PLUS"])
        )

    def test_parse_message_reject_long_keys_SCRAM(self):
        with self.assertRaisesRegexp(Exception, "protocol violation"):
            list(aiosasl.SCRAM.parse_message(b"foo=bar"))

    def test_parse_message_reject_long_keys_SCRAMPLUS(self):
        with self.assertRaisesRegexp(Exception, "protocol violation"):
            list(aiosasl.SCRAMPLUS.parse_message(b"foo=bar"))

    def test_parse_message_reject_m_key_SCRAM(self):
        with self.assertRaisesRegexp(Exception, "protocol violation"):
            list(aiosasl.SCRAM.parse_message(b"m=bar"))

    def test_parse_message_reject_m_key_SCRAMPLUS(self):
        with self.assertRaisesRegexp(Exception, "protocol violation"):
            list(aiosasl.SCRAMPLUS.parse_message(b"m=bar"))

    def test_parse_message_unescape_n_and_a_payload_SCRAM(self):
        data = list(aiosasl.SCRAM.parse_message(
            b"n=foo=2Cbar=3Dbaz,"
            b"a=fnord=2Cfunky=3Dfunk",
        ))
        self.assertSequenceEqual(
            [
                (b"n", b"foo,bar=baz"),
                (b"a", b"fnord,funky=funk")
            ],
            data
        )

    def test_parse_message_unescape_n_and_a_payload_SCRAMPLUS(self):
        data = list(aiosasl.SCRAMPLUS.parse_message(
            b"n=foo=2Cbar=3Dbaz,a=fnord=2Cfunky=3Dfunk"))
        self.assertSequenceEqual(
            [
                (b"n", b"foo,bar=baz"),
                (b"a", b"fnord,funky=funk")
            ],
            data
        )


class TestSCRAMImpl:
    def setUp(self):
        self.hashfun_factory = hashlib.sha1
        self.digest_size = self.hashfun_factory().digest_size
        self.user = b"user"
        self.user2 = "user\U0001f916".encode("utf-8")
        self.password = b"pencil"
        self.salt = b"QSXCR+Q6sek8bf92"

        aiosasl.scram._system_random = unittest.mock.MagicMock()
        aiosasl.scram._system_random.getrandbits.return_value = \
            int.from_bytes(b"foo", "little")

        self.salted_password = hashlib.pbkdf2_hmac(
            "sha1",
            self.password,
            self.salt,
            4096,
            self.digest_size)

        self.salted_password_4000 = hashlib.pbkdf2_hmac(
            "sha1",
            self.password,
            self.salt,
            4000,
            self.digest_size)

        self.salted_password_5000 = hashlib.pbkdf2_hmac(
            "sha1",
            self.password,
            self.salt,
            5000,
            self.digest_size)

        self.client_key = hmac.new(
            self.salted_password,
            b"Client Key",
            self.hashfun_factory).digest()

        self.client_key_4000 = hmac.new(
            self.salted_password_4000,
            b"Client Key",
            self.hashfun_factory).digest()

        self.client_key_5000 = hmac.new(
            self.salted_password_5000,
            b"Client Key",
            self.hashfun_factory).digest()

        self.stored_key = self.hashfun_factory(
            self.client_key).digest()

        self.stored_key_4000 = self.hashfun_factory(
            self.client_key_4000).digest()

        self.stored_key_5000 = self.hashfun_factory(
            self.client_key_5000).digest()

        self.client_first_message_bare = b"n=user,r=Zm9vAAAAAAAAAAAAAAAA"
        self.client_first_message_bare2 = \
            b"n="+self.user2+b",r=Zm9vAAAAAAAAAAAAAAAA"
        self.server_first_message = b"".join([
            b"r=Zm9vAAAAAAAAAAAAAAAA3rfcNHYJY1ZVvWVs7j,s=",
            base64.b64encode(self.salt),
            b",i=4096"
        ])
        self.server_first_message_4000 = b"".join([
            b"r=Zm9vAAAAAAAAAAAAAAAA3rfcNHYJY1ZVvWVs7j,s=",
            base64.b64encode(self.salt),
            b",i=4000"
        ])
        self.server_first_message_5000 = b"".join([
            b"r=Zm9vAAAAAAAAAAAAAAAA3rfcNHYJY1ZVvWVs7j,s=",
            base64.b64encode(self.salt),
            b",i=5000"
        ])

        if self._scram_plus == 'no':
            self.client_final_message_without_proof = (
                b"c=biws,r=Zm9vAAAAAAAAAAAAAAAA3rfcNHYJY1ZVvWVs7j")
        elif self._scram_plus == 'supported':
            self.client_final_message_without_proof = (
                b"c=eSws,r=Zm9vAAAAAAAAAAAAAAAA3rfcNHYJY1ZVvWVs7j")
        elif self._scram_plus == 'active':
            self.client_final_message_without_proof = (
                b"c=cD10bHMtdW5pcXVlLCxjaGFubmVsIGJpbmRpbmcgZGF0YQ==,"
                b"r=Zm9vAAAAAAAAAAAAAAAA3rfcNHYJY1ZVvWVs7j")
        else:
            raise Exception("invalid scram mode")

        self.auth_message = b",".join([
            self.client_first_message_bare,
            self.server_first_message,
            self.client_final_message_without_proof
        ])

        self.auth_message2 = b",".join([
            self.client_first_message_bare2,
            self.server_first_message,
            self.client_final_message_without_proof
        ])

        self.auth_message_4000 = b",".join([
            self.client_first_message_bare,
            self.server_first_message_4000,
            self.client_final_message_without_proof
        ])

        self.auth_message_5000 = b",".join([
            self.client_first_message_bare,
            self.server_first_message_5000,
            self.client_final_message_without_proof
        ])

        self.client_signature = hmac.new(
            self.stored_key,
            self.auth_message,
            self.hashfun_factory).digest()

        self.client_signature2 = hmac.new(
            self.stored_key,
            self.auth_message2,
            self.hashfun_factory).digest()

        self.client_signature_4000 = hmac.new(
            self.stored_key_4000,
            self.auth_message_4000,
            self.hashfun_factory).digest()

        self.client_signature_5000 = hmac.new(
            self.stored_key_5000,
            self.auth_message_5000,
            self.hashfun_factory).digest()

        self.client_proof = xor_bytes(self.client_signature, self.client_key)
        self.client_proof2 = xor_bytes(self.client_signature2, self.client_key)
        self.client_proof_4000 = xor_bytes(self.client_signature_4000,
                                           self.client_key_4000)
        self.client_proof_5000 = xor_bytes(self.client_signature_5000,
                                           self.client_key_5000)

        self.server_key = hmac.new(
            self.salted_password,
            b"Server Key",
            self.hashfun_factory).digest()
        self.server_key_4000 = hmac.new(
            self.salted_password_4000,
            b"Server Key",
            self.hashfun_factory).digest()
        self.server_key_5000 = hmac.new(
            self.salted_password_5000,
            b"Server Key",
            self.hashfun_factory).digest()
        self.server_signature = hmac.new(
            self.server_key,
            self.auth_message,
            self.hashfun_factory).digest()
        self.server_signature2 = hmac.new(
            self.server_key,
            self.auth_message2,
            self.hashfun_factory).digest()
        self.server_signature_4000 = hmac.new(
            self.server_key_4000,
            self.auth_message_4000,
            self.hashfun_factory).digest()
        self.server_signature_5000 = hmac.new(
            self.server_key_5000,
            self.auth_message_5000,
            self.hashfun_factory).digest()

        self._tls_connection = unittest.mock.Mock()
        self._tls_connection.get_finished = unittest.mock.Mock()
        self._tls_connection.get_finished.return_value = \
            b'channel binding data'

    async def _provide_credentials(self, *args):
        return ("user", "pencil")

    def _run(self, smmock, scram):
        info = aiosasl.scram.Base._supported_hashalgos["SHA-1"]
        if self._scram_plus in ('no', 'supported'):
            token = ("SCRAM-SHA-1", info)
        else:
            token = ("SCRAM-SHA-1-PLUS", info)

        result = asyncio.get_event_loop().run_until_complete(
            scram.authenticate(smmock, token)
        )
        smmock.interface.finalize()
        return result

    def tearDown(self):
        import random
        aiosasl.scram._system_random = random.SystemRandom()


class TestSCRAM(TestSCRAMImpl, unittest.TestCase):
    _scram_plus = 'no'

    def test_rfc(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof),
                 "success",
                 b"v="+base64.b64encode(self.server_signature))
            ]))

        self._run(
            smmock,
            aiosasl.SCRAM(self._provide_credentials)
        )

    def test_unassigned_password_codepoints(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            []))

        async def provide_credentials(*args):
            return ("user", "\U0001f916")

        with self.assertRaisesRegex(ValueError, "unassigned"):
            self._run(
                smmock,
                aiosasl.SCRAM(provide_credentials)
            )

    def test_unassigned_username_codepoints(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare2,
                 "challenge",
                 self.server_first_message),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof2),
                 "success",
                 b"v="+base64.b64encode(self.server_signature2))
            ]))

        async def provide_credentials(*args):
            return (self.user2.decode("utf-8"), self.password.decode("utf-8"))

        self._run(
            smmock,
            aiosasl.SCRAM(provide_credentials)
        )

    def test_malformed_reply(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare,
                 "challenge",
                 b"s=hut,t=hefu,c=kup,d=onny"),
                ("abort", None,
                 "failure", ("aborted", None))
            ]))

        with self.assertRaises(aiosasl.SASLFailure) as ctx:
            self._run(smmock, aiosasl.SCRAM(self._provide_credentials))

        self.assertIn(
            "malformed",
            str(ctx.exception).lower()
        )

    def test_other_malformed_reply(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare,
                 "challenge",
                 b"i=sometext,s=ABC,r=Zm9vAAAAAAAAAAAAAAAA3rfcNHYJY1ZVvWVs7j"),
                ("abort", None,
                 "failure", ("aborted", None))
            ]))

        with self.assertRaises(aiosasl.SASLFailure) as ctx:
            self._run(smmock, aiosasl.SCRAM(self._provide_credentials))

        self.assertIn(
            "malformed",
            str(ctx.exception).lower()
        )

    def test_incorrect_nonce(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare,
                 "challenge",
                 b"r=foobar,s="+base64.b64encode(self.salt)+b",i=4096"),
                ("abort", None,
                 "failure", ("aborted", None))
            ]))

        with self.assertRaisesRegexp(aiosasl.SASLFailure, "nonce") as ctx:
            self._run(smmock, aiosasl.SCRAM(self._provide_credentials))

        self.assertIsNone(ctx.exception.opaque_error)

    def test_invalid_signature(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof),
                 "success",
                 b"v="+base64.b64encode(b"fnord"))
            ]))

        with self.assertRaises(aiosasl.SASLFailure) as ctx:
            self._run(smmock, aiosasl.SCRAM(self._provide_credentials))

        self.assertIsNone(ctx.exception.opaque_error)
        self.assertIn(
            "signature",
            str(ctx.exception).lower()
        )

    def test_promote_failure_to_authentication_failure(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof),
                 "failure",
                 ("credentials-expired", None))
            ]))

        with self.assertRaises(aiosasl.AuthenticationFailure) as ctx:
            self._run(smmock, aiosasl.SCRAM(self._provide_credentials))

        self.assertEqual(
            "credentials-expired",
            ctx.exception.opaque_error
        )

    def test_reject_protocol_violation_1(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof),
                 "success",
                 None),
            ]))

        with self.assertRaisesRegexp(aiosasl.SASLFailure,
                                     "protocol violation") as ctx:
            self._run(smmock, aiosasl.SCRAM(self._provide_credentials))

        self.assertEqual(
            "malformed-request",
            ctx.exception.opaque_error
        )

    def test_reject_protocol_violation_2(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare,
                 "success", None),
                ("abort", None,
                 "failure", ("aborted", None)),
            ]))

        with self.assertRaisesRegexp(aiosasl.SASLFailure,
                                     "protocol violation") as ctx:
            self._run(smmock, aiosasl.SCRAM(self._provide_credentials))

        self.assertEqual(
            None,
            ctx.exception.opaque_error
        )

    def test_too_low_iteration_count(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message.replace(b",i=4096", b",i=4095")),
                ("abort", None,
                 "failure", ("aborted", None)),
            ]))

        with self.assertRaisesRegexp(
                aiosasl.SASLFailure,
                r"minimum iteration count for SCRAM-SHA-1 violated "
                r"\(4095 is less than 4096\)") as ctx:
            self._run(smmock, aiosasl.SCRAM(self._provide_credentials))

        self.assertEqual(
            None,
            ctx.exception.opaque_error
        )

    def test_too_low_iteration_count_without_enforcement(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message_4000),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof_4000),
                 "success",
                 b"v="+base64.b64encode(self.server_signature_4000))
            ]))

        self._run(
            smmock,
            aiosasl.SCRAM(
                self._provide_credentials,
                enforce_minimum_iteration_count=False,
            )
        )

    def test_high_iteration_count(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"n,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message_5000),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof_5000),
                 "success",
                 b"v="+base64.b64encode(self.server_signature_5000))
            ]))

        self._run(
            smmock,
            aiosasl.SCRAM(self._provide_credentials)
        )


class TestSCRAMDowngradeProtection(TestSCRAMImpl, unittest.TestCase):
    _scram_plus = 'supported'

    def test_rfc_with_downgrade_protection(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1",
                 b"y,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof),
                 "success",
                 b"v="+base64.b64encode(self.server_signature))
            ]))

        self._run(
            smmock,
            aiosasl.SCRAM(self._provide_credentials, after_scram_plus=True)
        )


class TestSCRAMPLUS(TestSCRAMImpl, unittest.TestCase):
    _scram_plus = 'active'

    def test_rfc(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1-PLUS",
                 b"p=tls-unique,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof),
                 "success",
                 b"v="+base64.b64encode(self.server_signature))
            ]))

        self._run(
            smmock,
            aiosasl.SCRAMPLUS(
                self._provide_credentials,
                TLSUnique(self._tls_connection)
            )
        )

    def test_malformed_reply(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1-PLUS",
                 b"p=tls-unique,,"+self.client_first_message_bare,
                 "challenge",
                 b"s=hut,t=hefu,c=kup,d=onny"),
                ("abort", None,
                 "failure", ("aborted", None))
            ]))

        with self.assertRaises(aiosasl.SASLFailure) as ctx:
            self._run(
                smmock,
                aiosasl.SCRAMPLUS(
                    self._provide_credentials,
                    TLSUnique(self._tls_connection)
                )
            )

        self.assertIn(
            "malformed",
            str(ctx.exception).lower()
        )

    def test_other_malformed_reply(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1-PLUS",
                 b"p=tls-unique,,"+self.client_first_message_bare,
                 "challenge",
                 b"i=sometext,s=ABC,r=Zm9vAAAAAAAAAAAAAAAA3rfcNHYJY1ZVvWVs7j"),
                ("abort", None,
                 "failure", ("aborted", None))
            ]))

        with self.assertRaises(aiosasl.SASLFailure) as ctx:
            self._run(
                smmock,
                aiosasl.SCRAMPLUS(
                    self._provide_credentials,
                    TLSUnique(self._tls_connection)
                )
            )

        self.assertIn(
            "malformed",
            str(ctx.exception).lower()
        )

    def test_incorrect_nonce(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1-PLUS",
                 b"p=tls-unique,,"+self.client_first_message_bare,
                 "challenge",
                 b"r=foobar,s="+base64.b64encode(self.salt)+b",i=4096"),
                ("abort", None,
                 "failure", ("aborted", None))
            ]))

        with self.assertRaisesRegexp(aiosasl.SASLFailure, "nonce") as ctx:
            self._run(
                smmock,
                aiosasl.SCRAMPLUS(
                    self._provide_credentials,
                    TLSUnique(self._tls_connection)
                )
            )

        self.assertIsNone(ctx.exception.opaque_error)

    def test_invalid_signature(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1-PLUS",
                 b"p=tls-unique,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof),
                 "success",
                 b"v="+base64.b64encode(b"fnord"))
            ]))

        with self.assertRaises(aiosasl.SASLFailure) as ctx:
            self._run(
                smmock,
                aiosasl.SCRAMPLUS(
                    self._provide_credentials,
                    TLSUnique(self._tls_connection)
                )
            )

        self.assertIsNone(ctx.exception.opaque_error)
        self.assertIn(
            "signature",
            str(ctx.exception).lower()
        )

    def test_promote_failure_to_authentication_failure(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1-PLUS",
                 b"p=tls-unique,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof),
                 "failure",
                 ("credentials-expired", None))
            ]))

        with self.assertRaises(aiosasl.AuthenticationFailure) as ctx:
            self._run(
                smmock,
                aiosasl.SCRAMPLUS(
                    self._provide_credentials,
                    TLSUnique(self._tls_connection)
                )
            )

        self.assertEqual(
            "credentials-expired",
            ctx.exception.opaque_error
        )

    def test_reject_protocol_violation(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;SCRAM-SHA-1-PLUS",
                 b"p=tls-unique,,"+self.client_first_message_bare,
                 "challenge",
                 self.server_first_message),
                ("response",
                 self.client_final_message_without_proof +
                 b",p="+base64.b64encode(self.client_proof),
                 "challenge",
                 b"foo"),
                ("response", b"", "success", b"bar")
            ]))

        with self.assertRaisesRegexp(aiosasl.SASLFailure,
                                     "protocol violation") as ctx:
            self._run(
                smmock,
                aiosasl.SCRAMPLUS(
                    self._provide_credentials,
                    TLSUnique(self._tls_connection)
                )
            )

        self.assertEqual(
            None,
            ctx.exception.opaque_error
        )


class TestANONYMOUS(unittest.TestCase):
    def test_accepts_ANONYMOUS(self):
        self.assertIsNotNone(
            aiosasl.ANONYMOUS.any_supported(["ANONYMOUS"])
        )

    def test_passes_token_through_trace(self):
        with unittest.mock.patch("aiosasl.stringprep.trace") as trace:
            trace.return_value = "traced"

            anon = aiosasl.ANONYMOUS(unittest.mock.sentinel.token)

        trace.assert_called_with(unittest.mock.sentinel.token)

        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;ANONYMOUS",
                 b"traced",
                 "success",
                 None)
            ]))

        async def run():
            await anon.authenticate(
                smmock,
                "ANONYMOUS")

        run_coroutine(run())

        smmock.interface.finalize()

    def test_simply_sends_token(self):
        smmock = aiosasl.SASLStateMachine(SASLInterfaceMock(
            self,
            [
                ("auth;ANONYMOUS",
                 b"sirhc",
                 "success",
                 None)
            ]))

        async def run():
            anon = aiosasl.ANONYMOUS("sirhc")
            await anon.authenticate(
                smmock,
                "ANONYMOUS")

        run_coroutine(run())

        smmock.interface.finalize()



#!/usr/bin/python3

import OpenSSL

key = OpenSSL.crypto.PKey()
key.generate_key(OpenSSL.crypto.TYPE_RSA, 1024)

encoded_certs = []
for algo, digest_algo in (("MD5", "SHA256"),
                          ("SHA1", "SHA256"),
                          ("SHA256", "SHA256"),
                          ("SHA512", "SHA512")):

    cert = OpenSSL.crypto.X509()
    cert.set_serial_number(1)
    cert.set_notBefore(b"20000101000000Z")
    cert.set_notAfter(b"21000101000000Z")
    cert.set_pubkey(key)
    cert.sign(key, algo)
    cert.get_subject().C = b"FN"
    cert.get_subject().O = b"Example Association"
    cert.get_subject().CN = b"aiosasl tests"
    cert.set_issuer(cert.get_subject())
    encoded_certs.append((
        OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_ASN1, cert),
        cert.digest(digest_algo)
    ))

print(encoded_certs)



########################################################################
# File name: __init__.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
Using SASL in a protocol
========================

To make use of SASL over an existing protocol, you first need to subclass and
implement :class:`SASLInterface`.

The usable mechanisms need to be detected by your application using the
protocol over which to implement SASL. This is generally protocol-specific. For
example, XMPP uses stream features to announce which SASL mechanisms are
supported by the server.

When a set of SASL mechanism strings has been obtained by the server (let us
call a set with the mechanism strings ``sasl_mechanisms``), the mechanisms
supported by your application (a list of :class:`SASLMechanism` subclass
instances, let us call it ``mechanism_impls``) can be queried for support::

    # intf = <instance of your subclass of SASLInterface>
    for impl in mechanism_impl:
        token = impl.any_supported(sasl_mechanisms)
        if token is not None:
            sm = aiosasl.SASLStateMachine(intf)
            try:
                await impl.authenticate(sm, token)
            except aiosasl.AuthenticationFailure:
                # handle authentication failure
                # it is generally not sensible to re-try with other mechanisms
            except aiosasl.SASLFailure:
                # this is a protocol problem, it is sensible to re-try other
                # mechanisms
            else:
                # authentication was successful!

The instances for the mechanisms can be re-used; they do not save any state,
the state is held by :class:`SASLStateMachine` instead. The different
mechanisms require different arguments (the password-based mechanisms generally
require a callback which provides credentials).

The mechanisms which are currently supported by :mod:`aiosasl` are summarised
below:

.. autosummary::

   ANONYMOUS
   PLAIN
   SCRAM
   SCRAMPLUS

Interface for protocols using SASL
==================================

To implement SASL on an existing protocol, you need to subclass
:class:`SASLInterface` and implement the abstract methods:

.. autoclass:: SASLInterface

.. autoclass:: SASLState

SASL mechansims
===============

.. autoclass:: PLAIN

.. autoclass:: SCRAM(credential_provider, *[, after_scram_plus=False][, enforce_minimum_iteration_count=True])

.. autoclass:: SCRAMPLUS(credential_provider, cb_provider, *[, enforce_minimum_iteration_count=True])

.. autoclass:: ANONYMOUS

Base class
----------

.. autoclass:: SASLMechanism

A note for implementers
-----------------------

The :class:`SASLStateMachine` unwraps `(SASLState.SUCCESS, payload)` messages
passed in from a :class:`SASLInterface` to the equivalent sequence
`(SASLState.CHALLENGE, payload)` (requiring the empty string as response) and
`(SASLState.SUCCESS, None)`. The two forms are equivalent as per the SASL
specification and this unwrapping allows uniform treatment of both
forms by the :class:`SASLMechanism` implementations.

SASL state machine
==================

.. autoclass:: SASLStateMachine

Exception classes
=================

.. autoclass:: SASLError

.. autoclass:: SASLFailure

.. autoclass:: AuthenticationFailure

Version information
===================

.. autodata:: __version__

.. autodata:: version_info
"""  # NOQA

from .common import (  # noqa:F401
    AuthenticationFailure,
    SASLError,
    SASLFailure,
    SASLState,
)

from .statemachine import (  # noqa:F401
    SASLInterface,
    SASLMechanism,
    SASLStateMachine,
)

from .scram import (  # noqa:F401
    SCRAM,
    SCRAMPLUS,
)

from .plain import (  # noqa:F401
    PLAIN,
)

from .anonymous import (  # noqa:F401
    ANONYMOUS,
)

from .version import version, __version__, version_info  # noqa:F401

#: The imported :mod:`aiosasl` version as a tuple.
#:
#: The components of the tuple are, in order: `major version`, `minor version`,
#: `patch level`, and `pre-release identifier`.
version_info = version_info

#: The imported :mod:`aiosasl` version as a string.
#:
#: The version number is dot-separated; in pre-release or development versions,
#: the version number is followed by a hypen-separated pre-release identifier.
__version__ = __version__



########################################################################
# File name: statemachine.py
# This file is part of: aiosasl
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import typing

from . import common


class SASLInterface(metaclass=abc.ABCMeta):
    """
    This class serves as an abstract base class for interfaces for use with
    :class:`SASLStateMachine`. Specific protocols using SASL (such as XMPP,
    IMAP or SMTP) can subclass this interface to implement SASL on top of the
    existing protocol.

    The interface class does not need to implement any state checking. State
    checking is done by the :class:`SASLStateMachine`. The following interface
    must be implemented by subclasses.

    The return values of the methods below are tuples of the following form:

    * ``(SASLState.SUCCESS, payload)`` -- After successful
      authentication, success is returned. Depending on the mechanism,
      a payload (as :class:`bytes` object) may be attached to the
      result, otherwise, ``payload`` is :data:`None`.

    * ``(SASLState.CHALLENGE, payload)`` -- A challenge was sent by
      the server in reply to the previous command.

    * ``(SASLState.FAILURE, None)`` -- This is only ever returned by
      :meth:`abort`. All other methods **must** raise errors as
      :class:`SASLFailure`.

    .. versionchanged:: 0.4

       The first element of the returned tuples are now elements of
       :class:`SASLState`. For compatibility with previous versions of
       ``aiosasl`` the first elements of the string may be one of the
       strings ``"success"``, ``"failure"`` or "``challenge``". For
       more information see :meth:`SASLState.from_reply`.

    .. automethod:: initiate

    .. automethod:: respond

    .. automethod:: abort
    """

    @abc.abstractmethod
    async def initiate(
            self,
            mechanism: str,
            payload: typing.Optional[bytes] = None,
            ) -> common.NextStateTuple:
        """
        Send a SASL initiation request for the given `mechanism`. Depending on
        the `mechanism`, an initial `payload` *may* be given. The `payload` is
        then a :class:`bytes` object which needs to be passed as initial
        payload during the initiation request.

        Wait for a reply by the peer and return the reply as a next-state tuple
        in the format documented at :class:`SASLInterface`.
        """

    @abc.abstractmethod
    async def respond(
            self,
            payload: bytes,
            ) -> common.NextStateTuple:
        """
        Send a response to a challenge. The `payload` is a :class:`bytes`
        object which is to be sent as response.

        Wait for a reply by the peer and return the reply as a next-state tuple
        in the format documented at :class:`SASLInterface`.
        """

    @abc.abstractmethod
    async def abort(self) -> None:
        """
        Abort the authentication. The result is either the failure tuple
        (``(SASLState.FAILURE, None)``) or a :class:`SASLFailure` exception if
        the response from the peer did not indicate abortion (e.g. another
        error was returned by the peer or the peer indicated success).
        """


class SASLStateMachine:
    """
    A state machine to reduce code duplication during SASL handshake.

    The state methods change the state and return the next client state of the
    SASL handshake, optionally with server-supplied payload.

    Note that, with the notable exception of :meth:`abort`, ``failure`` states
    are never returned but thrown as :class:`SASLFailure` instead.

    The initial state is never returned.
    """

    def __init__(self, interface: "SASLInterface"):
        super().__init__()
        self.interface = interface
        self._state = common.SASLState.INITIAL

    async def initiate(
            self,
            mechanism: str,
            payload: typing.Optional[bytes] = None,
            ) -> common.NextStateTuple:
        """
        Initiate the SASL handshake and advertise the use of the given
        `mechanism`. If `payload` is not :data:`None`, it will be base64
        encoded and sent as initial client response along with the ``<auth />``
        element.

        Return the next state of the state machine as tuple (see
        :class:`SASLStateMachine` for details).
        """

        if self._state != common.SASLState.INITIAL:
            raise RuntimeError("initiate has already been called")

        try:
            next_state, payload = await self.interface.initiate(
                mechanism,
                payload=payload)
        except common.SASLFailure:
            self._state = common.SASLState.FAILURE
            raise

        next_state = common.SASLState.from_reply(next_state)
        self._state = next_state
        return next_state, payload

    async def response(
            self,
            payload: bytes,
            ) -> common.NextStateTuple:
        """
        Send a response to the previously received challenge, with the given
        `payload`. The payload is encoded using base64 and transmitted to the
        server.

        Return the next state of the state machine as tuple (see
        :class:`SASLStateMachine` for details).
        """
        if self._state == common.SASLState.SUCCESS_SIMULATE_CHALLENGE:
            if payload != b"":
                # XXX: either our mechanism is buggy or the server
                # sent SASLState.SUCCESS before all challenge-response
                # messages defined by the mechanism were sent
                self._state = common.SASLState.FAILURE
                raise common.SASLFailure(
                    None,
                    "protocol violation: mechanism did not"
                    " respond with an empty response to a"
                    " challenge with final data – this suggests"
                    " a protocol-violating early success from the server."
                )
            self._state = common.SASLState.SUCCESS
            return common.SASLState.SUCCESS, None

        if self._state != common.SASLState.CHALLENGE:
            raise RuntimeError(
                "no challenge has been made or negotiation failed")

        try:
            next_state, response_payload = await self.interface.respond(
                payload,
            )
        except common.SASLFailure:
            self._state = common.SASLState.FAILURE
            raise

        next_state = common.SASLState.from_reply(next_state)

        # unfold the (SASLState.SUCCESS, payload) to a sequence of
        # (SASLState.CHALLENGE, payload), (SASLState.SUCCESS, None) for the
        # SASLMethod to allow uniform treatment of both cases
        if (next_state == common.SASLState.SUCCESS and
                response_payload is not None):
            self._state = common.SASLState.SUCCESS_SIMULATE_CHALLENGE
            return common.SASLState.CHALLENGE, response_payload

        self._state = next_state
        return next_state, response_payload

    async def abort(self) -> None:
        """
        Abort an initiated SASL authentication process. The expected result
        state is ``failure``.
        """
        if self._state == common.SASLState.INITIAL:
            raise RuntimeError("SASL authentication hasn't started yet")

        if self._state == common.SASLState.SUCCESS_SIMULATE_CHALLENGE:
            raise RuntimeError("SASL message exchange already over")

        try:
            return await self.interface.abort()
        finally:
            self._state = common.SASLState.FAILURE


class SASLMechanism(metaclass=abc.ABCMeta):
    """
    Implementation of a SASL mechanism. Two methods must be implemented by
    subclasses:

    .. automethod:: any_supported

    .. automethod:: authenticate

    .. note:: Administrative note

       Patches for new SASL mechanisms are welcome!

    """

    @abc.abstractclassmethod
    def any_supported(
            cls,
            mechanisms: typing.Iterable[str],
            ) -> typing.Any:
        """
        Determine whether this class can perform any SASL mechanism in the set
        of strings ``mechanisms``.

        If the class cannot perform any of the SASL mechanisms in
        ``mechanisms``, it must return :data:`None`.

        Otherwise, it must return a non-:data:`None` value. Applications must
        not assign any meaning to any value (except that :data:`None` is a sure
        indicator that the class cannot perform any of the listed mechanisms)
        and must not alter any value returned by this function. Note that even
        :data:`False` indicates success!

        The return value must be passed as second argument to
        :meth:`authenticate`. :meth:`authenticate` must not be called with a
        :data:`None` value.
        """

    async def authenticate(
            self,
            sm: SASLStateMachine,
            token: typing.Any,
            ) -> None:
        """
        Execute the mechanism identified by `token` (the non-:data:`None` value
        which has been returned by :meth:`any_supported` before) using the
        given :class:`SASLStateMachine` `sm`.

        If authentication fails, an appropriate exception is raised
        (:class:`AuthenticationFailure`). If the authentication fails for a
        reason unrelated to credentials, :class:`SASLFailure` is raised.
        """
